const MongoClient = require("mongodb").MongoClient;
const { Client } = require("pg");
const ttest = require("ttest");
const ObjectsToCsv = require("objects-to-csv");
const mysql = require("mysql");
const util = require("util");

const measuresSql = { csv: [] };
const measuresNosql = { csv: [] };
const measuresMongoSql = { csv: [] };

const measure = async (fn, id, globalMeasure) => {
  const res = [];
  for (let i = 0; i < 100; i++) {
    const start = process.hrtime.bigint();
    await fn();
    const end = process.hrtime.bigint();
    const diff = Number(end - start);
    res.push(Number(end - start));
    const csvRecord = globalMeasure.csv[i];
    globalMeasure.csv[i] = { ...csvRecord, [id]: diff };
  }
  globalMeasure[id] = res;
  console.log(
    id,
    res.reduce((p, c) => p + c, 0) / res.length,
    "in ms",
    res.reduce((p, c) => p + c, 0) / res.length / 1000000
  );
};

const noSqlPerformance = async () => {
  const dbName = "dipl";
  const noSql = new MongoClient("mongodb://localhost:27017", {
    useNewUrlParser: true
  });
  await noSql.connect();
  const db = noSql.db("dipl");
  const product = db.collection("product");
  const order = db.collection("order");
  const driveRecord = db.collection("drive_record");

  const productsOnDate = () =>
    order
      .aggregate([
        { $match: { date: new Date("2018-11-13T02:43:20.103Z") } },
        {
          $unwind: "$products"
        },
        {
          $lookup: {
            from: "product",
            localField: "products",
            foreignField: "_id",
            as: "joined_product"
          }
        },
        { $unwind: "$joined_product" },
        {
          $project: {
            _id: 1,
            date: 1,
            product_code: "$joined_product._id"
          }
        }
      ])
      .toArray();

  const orderDriveRecord = () =>
    driveRecord
      .aggregate([
        { $match: { _id: "drive_record2" } },
        {
          $project: {
            products: {
              $reduce: {
                input: "$orders.products",
                initialValue: [],
                in: { $setUnion: ["$$value", "$$this"] }
              }
            }
          }
        }
      ])
      .toArray();

  // console.log(JSON.stringify(await orderDriveRecord(), null, 4));

  const count = () => product.countDocuments();
  const sum = () =>
    product
      .aggregate([
        {
          $group: {
            _id: null,
            total: {
              $sum: "$price"
            }
          }
        }
      ])
      .next();
  const avg = () =>
    product
      .aggregate([
        {
          $group: {
            _id: null,
            total: {
              $avg: "$price"
            }
          }
        }
      ])
      .next();
  const min = () =>
    product
      .aggregate([
        {
          $group: {
            _id: null,
            total: {
              $min: "$price"
            }
          }
        }
      ])
      .next();
  const max = () =>
    product
      .aggregate([
        {
          $group: {
            _id: null,
            total: {
              $max: "$price"
            }
          }
        }
      ])
      .next();

  await measure(count, "count", measuresNosql);
  await measure(sum, "sum", measuresNosql);
  await measure(avg, "avg", measuresNosql);
  await measure(min, "min", measuresNosql);
  await measure(max, "max", measuresNosql);
  await measure(productsOnDate, "productsOnDate", measuresNosql);
  await measure(orderDriveRecord, "orderDriveRecord", measuresNosql);
  await noSql.close();
};

const sqlPerformance = async () => {
  const sql = new Client({
    user: "dipl",
    host: "localhost",
    database: "diplomka",
    password: "root",
    port: 5432
  });
  await sql.connect();
  const count = () => sql.query("select count(price) from product");
  const sum = () => sql.query("select sum(price) from product");
  const avg = () => sql.query("select avg(price) from product");
  const min = () => sql.query("select min(price) from product");
  const max = () => sql.query("select max(price) from product");
  const productsOnDate = () =>
    sql.query(`SELECT o.id, o.date, pr.product_code
  FROM public.order o
  JOIN order_product op ON (op.id_order = o.id)
  JOIN product pr ON (pr.id = op.id_product)
  WHERE o.date = '2019-02-21'
  ORDER BY o.id`);

  const orderDriveRecord = () =>
    sql.query(`SELECT DISTINCT po.id_product FROM drive_record dr
  JOIN order_drive_record odr ON (dr.id = odr.id_drive_record)
  JOIN order_product po on (odr.id = po.id_order)
  WHERE dr.id = 2`);

  await measure(count, "count", measuresSql);
  await measure(sum, "sum", measuresSql);
  await measure(avg, "avg", measuresSql);
  await measure(min, "min", measuresSql);
  await measure(max, "max", measuresSql);
  await measure(productsOnDate, "productsOnDate", measuresSql);
  await measure(orderDriveRecord, "orderDriveRecord", measuresSql);
  await sql.end();
};

const mongoSqlPerformance = async () => {
  const connection = mysql.createConnection({
    host: "localhost",
    port: 3307,
    user: "root",
    database: "dipl"
  });
  connection.query = util.promisify(connection.query);
  connection.connect = util.promisify(connection.connect);
  connection.end = util.promisify(connection.end);
  await connection.connect();
  const count = () => connection.query("select count(price) from product");
  const sum = () => connection.query("select sum(price) from product");
  const avg = () => connection.query("select avg(price) from product");
  const min = () => connection.query("select min(price) from product");
  const max = () => connection.query("select max(price) from product");

  const productsOnDate = () =>
    connection.query(`SELECT o._id, o.date, pr._id as pr_id
  FROM dipl.\`order\` o
  JOIN dipl.order_products op ON (op._id = o._id)
  JOIN dipl.product pr ON (pr._id = op.products)
  WHERE o.date >='2019-02-21 00:00:00'
  AND o.date <'2019-02-22 00:00:00'
  ORDER BY o._id`);

  const orderDriveRecord = () =>
    connection.query(`SELECT DISTINCT dro.\`orders.products\` FROM dipl.drive_record dr
    JOIN dipl.drive_record_orders_products dro ON (dr._id = dro._id)
    WHERE dr._id = "drive_record2" `);

  await measure(count, "count", measuresMongoSql);
  await measure(sum, "sum", measuresMongoSql);
  await measure(avg, "avg", measuresMongoSql);
  await measure(min, "min", measuresMongoSql);
  await measure(max, "max", measuresMongoSql);
  await measure(productsOnDate, "productsOnDate", measuresMongoSql);
  await measure(orderDriveRecord, "orderDriveRecord", measuresMongoSql);
  await connection.end();
};

module.exports = async multiplier => {
  console.log("//// PERF START ////");
  console.log("//// NOSQL ////");
  await noSqlPerformance();
  console.log("//// SQL ////");
  await sqlPerformance();
  console.log("//// SQL MONGO CONNECTOR ////");
  await mongoSqlPerformance();

  const ttestResult = [];

  Object.keys(measuresNosql).forEach((key, index) => {
    if (key !== "csv") {
      const res = ttest(measuresNosql[key], measuresSql[key], {
        alternative: "greater"
      });
      ttestResult.push({
        name: key,
        valid: JSON.stringify(res.valid()),
        confidence0: res.confidence()[0],
        confidence1: res.confidence()[1],
        pValue: res.pValue(),
        tValue: res.testValue()
      });
    }
  });
  await new ObjectsToCsv(measuresNosql.csv).toDisk(`nosql_${multiplier}.csv`);
  await new ObjectsToCsv(measuresSql.csv).toDisk(`sql_${multiplier}.csv`);
  await new ObjectsToCsv(measuresMongoSql.csv).toDisk(
    `mongoSql_${multiplier}.csv`
  );
  await new ObjectsToCsv(ttestResult.filter(val => val)).toDisk(
    `ttest_${multiplier}.csv`
  );

  console.log("//// PERF END ////");
};
